from __future__ import annotations
import re

from assemblyline.common.forge import CachedObject, get_datastore
from assemblyline.odm.models.tagging import Tagging


def tag_list_to_dict(tag_list: list[dict]) -> dict:
    tag_dict = {}
    for t in tag_list:
        if t['type'] not in tag_dict:
            tag_dict[t['type']] = []
        tag_dict[t['type']].append(t['value'])

    return tag_dict


def tag_dict_to_list(tag_dict: dict, safelisted: bool = False) -> list[dict]:
    return [
        {'safelisted': safelisted, 'type': k, 'value': t, 'short_type': k.rsplit(".", 1)[-1]}
        for k, v in tag_dict.items()
        if v is not None
        for t in v
    ]


def get_safelist_key(t_type: str, t_value: str) -> str:
    return f"{t_type}__{t_value}"


def get_safelist(ds) -> dict[str, bool]:
    return {get_safelist_key(sl['tag']['type'], sl['tag']['value']): True
            for sl in ds.safelist.stream_search("type:tag AND enabled:true", as_obj=False)}


class InvalidSafelist(Exception):
    pass


class TagSafelister(object):
    def __init__(self, data, log=None):
        valid_tags = set(Tagging.flat_fields().keys())
        self.datastore = get_datastore()
        self.safelist = CachedObject(get_safelist, kwargs={'ds': self.datastore}, refresh=300)

        self.match = data.get('match', {})
        self.regex = data.get('regex', {})
        self.log = log

        # Validate matches and regex
        for section, item in {'match': self.match, 'regex': self.regex}.items():
            if not isinstance(item, dict):
                raise InvalidSafelist(f"Section {section} should be of type: DICT")

            for k, v in item.items():
                if not isinstance(v, list):
                    raise InvalidSafelist(f"Values in the {section} section should all be of type: LIST")

                if k not in valid_tags:
                    raise InvalidSafelist(f"Key ({k}) in the {section} section is not a valid tag.")

                if section == 'regex':
                    self.regex[k] = [re.compile(x) for x in v]

    def is_safelisted(self, t_type, t_value):
        if self.safelist.get(get_safelist_key(t_type, t_value), False):
            if self.log:
                self.log.info(f"Tag '{t_type}' with value '{t_value}' was safelisted.")
            return True

        for match in self.match.get(t_type, []):
            if t_value == match:
                if self.log:
                    self.log.info(f"Tag '{t_type}' with value '{t_value}' was safelisted by match rule.")
                return True

        for regex in self.regex.get(t_type, []):
            if regex.match(t_value):
                if self.log:
                    self.log.info(f"Tag '{t_type}' with value '{t_value}' "
                                  f"was safelisted by regex '{regex.pattern}'.")
                return True

        return False

    def safelist_many(self, t_type, t_values):
        if not isinstance(t_values, list):
            t_values = [t_values]

        tags = []
        safelisted_tags = []
        for x in t_values:
            if self.is_safelisted(t_type, x):
                safelisted_tags.append(x)
            else:
                tags.append(x)

        return tags, safelisted_tags

    def get_validated_tag_map(self, tag_map):
        tags = {}
        safelisted_tags = {}
        for k, v in tag_map.items():
            if v is not None and v != []:
                c_tags, c_safelisted_tags = self.safelist_many(k, v)
                if c_tags:
                    tags[k] = c_tags
                if c_safelisted_tags:
                    safelisted_tags[k] = c_safelisted_tags

        return tags, safelisted_tags




hostname = 'unknownhost'
# noinspection PyBroadException
try:
    from assemblyline.common.net import get_hostname
    hostname = get_hostname()
except Exception:  # pylint:disable=W0702
    pass

ip = 'x.x.x.x'
# noinspection PyBroadException
try:
    from assemblyline.common.net import get_hostip
    ip = get_hostip()
except Exception:  # pylint:disable=W0702
    pass

AL_SYSLOG_FORMAT = f'{ip} AL %(levelname)8s %(process)5d %(name)40s | %(message)s'
AL_LOG_FORMAT = f'%(asctime)-16s %(levelname)8s {hostname} %(process)d %(name)40s | %(message)s'
AL_JSON_FORMAT = f'{{' \
    f'"@timestamp": "%(asctime)s", ' \
    f'"event": {{ "module": "assemblyline", "dataset": "%(name)s" }}, ' \
    f'"host": {{ "ip": "{ip}", "hostname": "{hostname}" }}, ' \
    f'"log": {{ "level": "%(levelname)s", "logger": "%(name)s" }}, ' \
    f'"process": {{ "pid": "%(process)d" }}, ' \
    f'"message": %(message)s}}'



from assemblyline import odm


@odm.model(index=True, store=True, description="File Seen Model")
class Seen(odm.Model):
    count = odm.Integer(default=1, description="How many times have we seen this file?")
    first = odm.Date(default="NOW", description="First seen timestamp")
    last = odm.Date(default="NOW", description="Last seen timestamp")


@odm.model(index=True, store=True, description="Model of File")
class File(odm.Model):
    archive_ts = odm.Optional(odm.Date(store=False, description="Archiving timestamp (Deprecated)"))
    ascii = odm.Keyword(index=False, store=False,
                        description="Dotted ASCII representation of the first 64 bytes of the file")
    classification = odm.Classification(description="Classification of the file")
    entropy = odm.Float(description="Entropy of the file")
    expiry_ts = odm.Optional(odm.Date(store=False), description="Expiry timestamp")
    is_section_image = odm.Boolean(default=False, description="Is this an image from an Image Result Section?")
    hex = odm.Keyword(index=False, store=False, description="Hex dump of the first 64 bytes of the file")
    md5 = odm.MD5(copyto="__text__", description="MD5 of the file")
    magic = odm.Keyword(store=False, description="Output from libmagic related to the file")
    mime = odm.Optional(odm.Keyword(store=False), description="MIME type of the file as identified by libmagic")
    seen = odm.Compound(Seen, default={}, description="Details about when the file was seen")
    sha1 = odm.SHA1(copyto="__text__", description="SHA1 hash of the file")
    sha256 = odm.SHA256(copyto="__text__", description="SHA256 hash of the file")
    size = odm.Integer(description="Size of the file in bytes")
    ssdeep = odm.SSDeepHash(store=False, description="SSDEEP hash of the file")
    type = odm.Keyword(copyto="__text__", description="Type of file as identified by Assemblyline")
    from_archive = odm.Boolean(index=False, default=False, description="Was loaded from the archive")



from __future__ import annotations

import logging
import re
import time
import typing

from os import environ
from random import random
from urllib.parse import urlparse

import elasticsearch
import elasticsearch.helpers

from assemblyline.common import forge
from assemblyline.datastore.collection import ESCollection
from assemblyline.datastore.exceptions import (DataStoreException, UnsupportedElasticVersion,
                                               SearchRetryException, VersionConflictException)

from packaging import version

TRANSPORT_TIMEOUT = int(environ.get('AL_DATASTORE_TRANSPORT_TIMEOUT', '90'))
log = logging.getLogger('assemblyline.datastore')


class ESStore(object):
    """ Elasticsearch multi-index implementation of the ResultStore interface."""
    MAX_RETRY_BACKOFF = 10
    DEFAULT_SORT = "id asc"
    DATE_FORMAT = {
        'NOW': 'now',
        'YEAR': 'y',
        'MONTH': 'M',
        'WEEK': 'w',
        'DAY': 'd',
        'HOUR': 'h',
        'MINUTE': 'm',
        'SECOND': 's',
        'MILLISECOND': 'ms',
        'MICROSECOND': 'micros',
        'NANOSECOND': 'nanos',
        'SEPARATOR': '||',
        'DATE_END': 'Z'
    }
    DATEMATH_MAP = {
        'NOW': 'now',
        'YEAR': 'y',
        'MONTH': 'M',
        'WEEK': 'w',
        'DAY': 'd',
        'HOUR': 'h',
        'MINUTE': 'm',
        'SECOND': 's',
        'DATE_END': 'Z||'
    }
    ID = 'id'
    MIN_ELASTIC_VERSION = '7.10'

    def __init__(self, hosts, archive_access=True):
        config = forge.get_config()
        self._hosts = hosts
        self._closed = False
        self._collections = {}
        self._models = {}
        self.archive_indices = config.datastore.archive.indices if config.datastore.archive.enabled else []
        self.validate = True

        tracer = logging.getLogger('elasticsearch')
        tracer.setLevel(logging.CRITICAL)

        self.client = elasticsearch.Elasticsearch(hosts=hosts,
                                                  max_retries=0,
                                                  request_timeout=TRANSPORT_TIMEOUT)
        self.es_version = version.parse(self.client.info()['version']['number'])
        self.archive_access = archive_access
        self.url_path = 'elastic'
        self._test_elastic_minimum_version()

    def __enter__(self):
        return self

    def __exit__(self, ex_type, exc_val, exc_tb):
        self.close()

    def __str__(self):
        return '{0} - {1}'.format(self.__class__.__name__, self._hosts)

    def __getattr__(self, name) -> ESCollection:
        if not self.validate:
            return ESCollection(self, name, model_class=self._models[name], validate=self.validate)

        if name not in self._collections:
            self._collections[name] = ESCollection(
                self, name, model_class=self._models[name], validate=self.validate)

        return self._collections[name]

    @property
    def now(self):
        return self.DATE_FORMAT['NOW']

    @property
    def ms(self):
        return self.DATE_FORMAT['MILLISECOND']

    @property
    def us(self):
        return self.DATE_FORMAT['MICROSECOND']

    @property
    def ns(self):
        return self.DATE_FORMAT['NANOSECOND']

    @property
    def year(self):
        return self.DATE_FORMAT['YEAR']

    @property
    def month(self):
        return self.DATE_FORMAT['MONTH']

    @property
    def week(self):
        return self.DATE_FORMAT['WEEK']

    @property
    def day(self):
        return self.DATE_FORMAT['DAY']

    @property
    def hour(self):
        return self.DATE_FORMAT['HOUR']

    @property
    def minute(self):
        return self.DATE_FORMAT['MINUTE']

    @property
    def second(self):
        return self.DATE_FORMAT['SECOND']

    @property
    def date_separator(self):
        return self.DATE_FORMAT['SEPARATOR']

    def _test_elastic_minimum_version(self):
        if not self.is_supported_version(self.MIN_ELASTIC_VERSION):
            raise UnsupportedElasticVersion(f"Elastic version {self.es_version} is not supported by Assemblyline. "
                                            f"Upgrade to Elastic {self.MIN_ELASTIC_VERSION} at minimum.")

    def is_supported_version(self, min):
        return self.es_version >= version.parse(min)

    def connection_reset(self):
        self.client = elasticsearch.Elasticsearch(hosts=self._hosts,
                                                  max_retries=0,
                                                  request_timeout=TRANSPORT_TIMEOUT)
        self.es_version = version.parse(self.client.info()['version']['number'])
        self._test_elastic_minimum_version()

    def close(self):
        self._closed = True
        # Flatten the client object so that attempts to access without reconnecting errors hard
        # But 'cast' it so that mypy and other linters don't think that its normal for client to be None
        self.client = typing.cast(elasticsearch.Elasticsearch, None)

    def get_hosts(self, safe=False):
        if not safe:
            return self._hosts
        else:
            out = []
            for h in self._hosts:
                parsed = urlparse(h)
                out.append(parsed.hostname or parsed.path)
            return out

    def get_models(self):
        return self._models

    def is_closed(self):
        return self._closed

    def ping(self):
        return self.client.ping()

    def register(self, name: str, model_class=None):
        name_match = re.match(r'[a-z0-9_]*', name)
        if not name_match or name_match.string != name:
            raise DataStoreException('Invalid characters in model name. '
                                     'You can only use lower case letters, numbers and underscores.')

        self._models[name] = model_class

    def to_pydatemath(self, value):
        replace_list = [
            (self.now, self.DATEMATH_MAP['NOW']),
            (self.year, self.DATEMATH_MAP['YEAR']),
            (self.month, self.DATEMATH_MAP['MONTH']),
            (self.week, self.DATEMATH_MAP['WEEK']),
            (self.day, self.DATEMATH_MAP['DAY']),
            (self.hour, self.DATEMATH_MAP['HOUR']),
            (self.minute, self.DATEMATH_MAP['MINUTE']),
            (self.second, self.DATEMATH_MAP['SECOND']),
            (self.DATE_FORMAT['DATE_END'], self.DATEMATH_MAP['DATE_END'])
        ]

        for x in replace_list:
            value = value.replace(*x)

        return value

    def with_retries(self, func, *args, raise_conflicts=False, **kwargs):
        """
        This function performs the passed function with the given args and kwargs and reconnect if it fails

        :return: return the output of the function passed
        """
        retries = 0
        updated = 0
        deleted = 0
        while True:
            try:
                ret_val = func(*args, **kwargs)

                if retries:
                    log.info('Reconnected to elasticsearch!')

                if updated:
                    ret_val['updated'] += updated

                if deleted:
                    ret_val['deleted'] += deleted

                return ret_val

            except elasticsearch.exceptions.ConflictError as ce:
                if raise_conflicts:
                    # De-sync potential treads trying to write to the index
                    time.sleep(random() * 0.1)
                    raise VersionConflictException(str(ce))
                updated += ce.info.get('updated', 0)
                deleted += ce.info.get('deleted', 0)

                time.sleep(min(retries, self.MAX_RETRY_BACKOFF))
                self.connection_reset()
                retries += 1

            except elasticsearch.exceptions.ConnectionTimeout:
                log.warning(f"Elasticsearch connection timeout, server(s): "
                            f"{' | '.join(self.get_hosts(safe=True))}"
                            f", retrying {func.__name__}...")
                time.sleep(min(retries, self.MAX_RETRY_BACKOFF))
                self.connection_reset()
                retries += 1

            except (SearchRetryException,
                    elasticsearch.exceptions.ConnectionError,
                    elasticsearch.exceptions.AuthenticationException) as e:
                if not isinstance(e, SearchRetryException):
                    log.warning(f"No connection to Elasticsearch server(s): "
                                f"{' | '.join(self.datastore.get_hosts(safe=True))}"
                                f", because [{e}] retrying {func.__name__}...")

                time.sleep(min(retries, self.MAX_RETRY_BACKOFF))
                self.connection_reset()
                retries += 1

            except elasticsearch.exceptions.TransportError as e:
                err_code, _, _ = e.args
                if err_code == 503 or err_code == '503':
                    log.warning(f"Looks like index {self.name} is not ready yet, retrying...")
                    time.sleep(min(retries, self.MAX_RETRY_BACKOFF))
                    self.connection_reset()
                    retries += 1
                elif err_code == 429 or err_code == '429':
                    log.warning("Elasticsearch is too busy to perform the requested "
                                f"task on index {self.name}, retrying...")
                    time.sleep(min(retries, self.MAX_RETRY_BACKOFF))
                    self.connection_reset()
                    retries += 1
                elif err_code == 403 or err_code == '403':
                    log.warning("Elasticsearch cluster is preventing writing operations "
                                f"on index {self.name}, retrying...")
                    time.sleep(min(retries, self.MAX_RETRY_BACKOFF))
                    self.connection_reset()
                    retries += 1

                else:
                    raise




from assemblyline import odm
from assemblyline.common.caching import generate_conf_key

STATUSES = {"FAIL_NONRECOVERABLE", "FAIL_RECOVERABLE"}
ERROR_TYPES = {
    "UNKNOWN": 0,
    "EXCEPTION": 1,
    "MAX DEPTH REACHED": 10,
    "MAX FILES REACHED": 11,
    "MAX RETRY REACHED": 12,
    "SERVICE BUSY": 20,
    "SERVICE DOWN": 21,
    "TASK PRE-EMPTED": 30
}


@odm.model(index=True, store=True, description="Error Response from a Service")
class Response(odm.Model):
    message = odm.Text(copyto="__text__", description="Error message")
    service_debug_info = odm.Optional(odm.Keyword(), description="Information about where the service was processed")
    service_name = odm.Keyword(copyto="__text__", description="Service Name")
    service_tool_version = odm.Optional(odm.Keyword(copyto="__text__"), description="Service Tool Version")
    service_version = odm.Keyword(description="Service Version")
    status = odm.Enum(values=STATUSES, description="Status of error produced by service")


@odm.model(index=True, store=True, description="Error Model used by Error Viewer")
class Error(odm.Model):
    archive_ts = odm.Optional(odm.Date(store=False, description="Archiving timestamp (Deprecated)"))
    created = odm.Date(default="NOW", description="Error creation timestamp")
    expiry_ts = odm.Optional(odm.Date(store=False), description="Expiry timestamp")
    response: Response = odm.Compound(Response, description="Response from the service")
    sha256 = odm.SHA256(copyto="__text__", description="SHA256 of file related to service error")
    type = odm.Enum(values=list(ERROR_TYPES.keys()), default="EXCEPTION", description="Type of error")

    def build_key(self, service_tool_version=None, task=None):
        key_list = [
            self.sha256,
            self.response.service_name.replace('.', '_'),
            f"v{self.response.service_version.replace('.', '_')}",
            f"c{generate_conf_key(service_tool_version=service_tool_version, task=task)}",
            f"e{ERROR_TYPES.get(self.type, 0)}"]

        return '.'.join(key_list)



import boto3
import logging
import os
import tempfile
from typing import Iterable, Optional

from botocore.exceptions import ClientError, EndpointConnectionError, ConnectionClosedError
from io import BytesIO

from assemblyline.common.exceptions import ChainAll
from assemblyline.filestore.transport.base import Transport, TransportException

try:
    from botocore.vendored.requests.packages.urllib3 import disable_warnings
except ImportError:
    from urllib3 import disable_warnings


disable_warnings()

"""
This class assumes a flat file structure in the S3 bucket.  This is due to the way the AL datastore currently handles
file paths for local/ftp datastores not playing nicely with s3 constraints.
"""


@ChainAll(TransportException)
class TransportS3(Transport):
    DEFAULT_HOST = "s3.amazonaws.com"

    def __init__(self, base=None, accesskey=None, secretkey=None, aws_region=None, s3_bucket="al-storage",
                 host=None, port=None, use_ssl=None, verify=True, connection_attempts=None):
        self.log = logging.getLogger('assemblyline.transport.s3')
        self.base = base
        self.bucket = s3_bucket
        self.accesskey = accesskey
        self.retry_limit: int = connection_attempts

        if use_ssl is None:
            self.use_ssl = True
        else:
            self.use_ssl = use_ssl

        if host is None:
            self.host = self.DEFAULT_HOST
        else:
            self.host = host

        if port is None:
            self.port = {True: 443, False: 80}[self.use_ssl]
        else:
            self.port = port

        self.scheme = {True: "https", False: "http"}[self.use_ssl]

        self.endpoint_url = "{scheme}://{host}:{port}".format(scheme=self.scheme, host=self.host, port=self.port)

        session = boto3.session.Session()
        self.client = session.client(
            "s3",
            aws_access_key_id=accesskey,
            aws_secret_access_key=secretkey,
            endpoint_url=self.endpoint_url,
            region_name=aws_region,
            use_ssl=self.use_ssl,
            verify=verify
        )

        bucket_exist = False
        try:
            self.with_retries(self.client.head_bucket, Bucket=self.bucket)
            bucket_exist = True
        except TransportException as e:
            if isinstance(e.cause, ClientError) and e.cause.response['Error']['Code'] == '404':
                pass
            else:
                raise

        if not bucket_exist:
            try:
                self.with_retries(self.client.create_bucket, Bucket=self.bucket)
            except TransportException as e:
                if isinstance(e.cause, ClientError) and e.cause.response['Error']['Code'] == "BucketAlreadyOwnedByYou":
                    # By the time that we listed the bucket an found it didn't exist, someone else created it.
                    pass
                else:
                    raise

        def s3_normalize(path):
            # flatten path to just the basename
            return os.path.basename(path)

        super(TransportS3, self).__init__(normalize=s3_normalize)

    def __str__(self):
        out = "s3://"
        if self.accesskey:
            out += "%s@" % self.accesskey
        out += "%s:%s" % (self.host, self.port)
        if self.bucket:
            out += "/%s" % self.bucket
        if self.base:
            out += self.base
        return out

    def with_retries(self, func, *args, **kwargs):
        retries = 0
        while self.retry_limit is None or retries <= self.retry_limit:
            try:
                ret_val = func(*args, **kwargs)

                if retries:
                    self.log.info('Reconnected to S3 transport!')

                return ret_val

            except (EndpointConnectionError, ConnectionClosedError):
                self.log.warning(f"No connection to S3 transport {self.endpoint_url}, retrying...")
                retries += 1
        raise ConnectionError(f"Couldn't connect to the requested S3 endpoint {self.endpoint_url} inside retry limit")

    def delete(self, path):
        key = self.normalize(path)
        self.with_retries(self.client.delete_object, Bucket=self.bucket, Key=key)

    def exists(self, path):
        # checks to see if KEY exists
        key = self.normalize(path)
        self.log.debug('Checking for existence of %s', key)
        try:
            self.with_retries(self.client.head_object, Bucket=self.bucket, Key=key)
        except TransportException as error:
            if isinstance(error.cause, ClientError):
                return False
            raise
        return True

    def makedirs(self, path):
        # Does not need to do anything as s3 has a flat layout.
        pass

    # File based functions
    def download(self, src_path, dst_path):
        key = self.normalize(src_path)
        dir_path = os.path.dirname(dst_path)
        # create dst_path if it doesn't exist
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)
        # download the key from s3
        self.with_retries(self.client.download_file, self.bucket, key, dst_path)

    def upload(self, src_path, dst_path):
        dst_path = self.normalize(dst_path)
        # if file exists already, it will be overwritten
        self.with_retries(self.client.upload_file, src_path, self.bucket, dst_path)

    # Buffer based functions
    def get(self, path: str) -> bytes:
        fd, dst_path = tempfile.mkstemp(prefix="s3_transport.", suffix=".download")
        os.close(fd)  # We don't need the file descriptor open

        self.with_retries(self.download, path, dst_path)
        try:
            with open(dst_path, "rb") as downloaded:
                return downloaded.read()
        finally:
            if os.path.exists(dst_path):
                os.remove(dst_path)

    def put(self, dst_path, content):
        dst_path = self.normalize(dst_path)
        if isinstance(content, str):
            content = content.encode('utf-8')

        with BytesIO(content) as file_io:
            self.with_retries(self.client.upload_fileobj, file_io, self.bucket, dst_path)

    def list(self, prefix: Optional[str] = None) -> Iterable[str]:
        args = {
            'Bucket': self.bucket,
            'Prefix': prefix or '',
            'MaxKeys': 50000,
        }
        while args.get('ContinuationToken', None) != '':
            data = self.client.list_objects_v2(**args)
            args['ContinuationToken'] = data.get("NextContinuationToken", '')
            for chunk in data.get('Contents', []):
                yield chunk['Key']



"""
Helper functions for the CLI to use in garbage collecting the filestore.
"""
from __future__ import annotations
import multiprocessing
import ctypes
from queue import Empty

from assemblyline.common.forge import get_datastore
from assemblyline.filestore import create_transport


FILTER_WORKERS = 8
DELETE_WORKERS = 8


def list_files(transport_url: str, listing_running, listed_queue, listing_count):
    """"A worker that feeds a queue with all the files listed by a filestore."""
    total_listed = 0
    try:
        # Connect to the transport and send the file names, count how many are sent
        transport = create_transport(transport_url)
        for filename in transport.list():
            total_listed += 1
            listed_queue.put(filename)
    finally:
        # Save the number sent to shared memory for other processes to read
        with listing_count.get_lock():
            listing_count.value = total_listed
        # Mark in shared memory that this worker is terminating
        with listing_running.get_lock():
            listing_running.value = 0


def filter_files(listing_running, listed_queue, filtering_running, filtered_queue, filtering_count):
    """
    A worker that reads the listed files, and filters out those that have
    entries in the file or cached_file indices of the database.

    Files without entries are fed into an output queue.
    """
    total_filtered = 0
    try:
        datastore = get_datastore(archive_access=True)
        while True:
            # Get a filename from the listing worker
            try:
                filename = listed_queue.get(timeout=3)
            except Empty:
                # If the queue is empty, check if the listing worker is finished
                with listing_running.get_lock():
                    if listing_running.value == 0:
                        break
                continue

            # Check if the file exists in the file index
            if datastore.file.get_if_exists(filename, archive_access=True):
                total_filtered += 1
                continue

            # Check if the file exists in the cached file index
            if datastore.cached_file.get_if_exists(filename, archive_access=True):
                total_filtered += 1
                continue

            # File has no records in datastore, feed to erase worker
            filtered_queue.put(filename)
    finally:
        # Add the number of files processed by this worker to a shared memory value
        with filtering_count.get_lock():
            filtering_count.value += total_filtered
        # Decrease the counter of running filter workers to account for this worker
        # exiting.
        with filtering_running.get_lock():
            filtering_running.value -= 1


def erase_files(transport_url, filtering_running, filtered_queue, erase_count):
    """
    A worker that reads from the filtered files queue and erases the
    specified files from the filestore.
    """
    total_erased = 0
    try:
        transport = create_transport(transport_url)
        while True:
            # Read files from the filtered file queue, exit if no more filter
            # workers are running
            try:
                filename = filtered_queue.get(timeout=3)
            except Empty:
                with filtering_running.get_lock():
                    if filtering_running.value == 0:
                        break
                continue

            # Erase the file fed to this worker
            transport.delete(filename)
            total_erased += 1
    finally:
        # Add the number of files erased by this worker to a shared memory worker
        with erase_count.get_lock():
            erase_count.value += total_erased


def cleanup_filestore(transport_url: str) -> str:
    # Create and initialize shared memory values and queues to communicate between the workers
    listing_count = multiprocessing.Value(ctypes.c_ulonglong)
    listing_count.value = 0
    listing_running = multiprocessing.Value(ctypes.c_ulonglong)
    listing_running.value = 1
    listed_queue = multiprocessing.Queue(10000)

    filtering_count = multiprocessing.Value(ctypes.c_ulonglong)
    filtering_count.value = 0
    filtering_running = multiprocessing.Value(ctypes.c_ulonglong)
    filtering_running.value = FILTER_WORKERS
    filtered_queue = multiprocessing.Queue(50000)

    erase_count = multiprocessing.Value(ctypes.c_ulonglong)
    erase_count.value = 0

    # Spawn process to list files
    listing_worker = multiprocessing.Process(
        target=list_files,
        args=(transport_url, listing_running, listed_queue, listing_count),
        daemon=True
    )
    listing_worker.start()

    # Spawn workers to filter file names
    filter_workers = [
        multiprocessing.Process(
            target=filter_files,
            args=(listing_running, listed_queue, filtering_running, filtered_queue, filtering_count),
            daemon=True
        )
        for _ in range(FILTER_WORKERS)
    ]
    for _f in filter_workers:
        _f.start()

    # Spawn workers to erase files
    erase_workers = [
        multiprocessing.Process(
            target=erase_files,
            args=(transport_url, filtering_running, filtered_queue, erase_count),
            daemon=True
        )
        for _ in range(DELETE_WORKERS)
    ]
    for _f in erase_workers:
        _f.start()

    # Wait for work to finish in order
    listing_worker.join()
    for _f in filter_workers:
        _f.join()
    for _f in erase_workers:
        _f.join()

    # Return a status string for the CLI
    return f'Files listed: {listing_count.value:,}; filtered: {filtering_count.value:,}; erased: {erase_count.value:,}'



from assemblyline import odm


@odm.model(index=True, store=False, description="Tagging Model")
class Tagging(odm.Model):
    @odm.model(index=True, store=False, description="Attribution Tag Model")
    class Attribution(odm.Model):
        actor = odm.Optional(odm.List(odm.UpperKeyword(copyto="__text__")), description="Attribution Actor")
        campaign = odm.Optional(odm.List(odm.UpperKeyword(copyto="__text__")), description="Attribution Campaign")
        exploit = odm.Optional(odm.List(odm.UpperKeyword(copyto="__text__")), description="Attribution Exploit")
        implant = odm.Optional(odm.List(odm.UpperKeyword(copyto="__text__")), description="Attribution Implant")
        family = odm.Optional(odm.List(odm.UpperKeyword(copyto="__text__")), description="Attribution Family")
        network = odm.Optional(odm.List(odm.UpperKeyword(copyto="__text__")), description="Attribution Network")

    @odm.model(index=True, store=False, description="Antivirus Tag Model")
    class AV(odm.Model):
        heuristic = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="List of heuristics")
        virus_name = odm.Optional(odm.List(odm.Keyword(copyto="__text__")),
                                  description="Collection of virus names identified by antivirus tools")

    @odm.model(index=True, store=False, description="Certificate Tag Model")
    class Cert(odm.Model):
        @odm.model(index=True, store=False, description="Valid Certificate Period")
        class CertValid(odm.Model):
            start = odm.Optional(odm.List(odm.Keyword(copyto="__text__")),
                                 description="Start date of certificate validity")
            end = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="End date of certificate validity")

        extended_key_usage = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Extended Key Usage")
        issuer = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Issuer")
        key_usage = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Key Usage")
        owner = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Owner")
        serial_no = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Serial Number")
        signature_algo = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Signature Algorithm")
        subject = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Subject Name")
        subject_alt_name = odm.Optional(
            odm.List(odm.Keyword(copyto="__text__")),
            description="Alternative Subject Name")
        thumbprint = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Thumbprint")
        valid = odm.Optional(odm.Compound(CertValid), description="Validity Information")
        version = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Version")

    @odm.model(index=True, store=False, description="Dynamic Tag Model. Commonly Used by Dynamic Analysis")
    class Dynamic(odm.Model):
        @odm.model(index=True, store=False, description="Dynamic Process")
        class DynamicProcess(odm.Model):
            command_line = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Commandline")
            file_name = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Filename")
            shortcut = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Shortcut")

        @odm.model(index=True, store=False, description="Signatures")
        class DynamicSignature(odm.Model):
            category = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Signature Category")
            family = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Signature Family")
            name = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Signature Name")

        @odm.model(index=True, store=False, description="SSDeep")
        class DynamicSSDeep(odm.Model):
            cls_ids = odm.Optional(odm.List(odm.SSDeepHash(copyto="__text__")), description="CLSIDs")
            dynamic_classes = odm.Optional(odm.List(odm.SSDeepHash(copyto="__text__")), description="Dynamic Classes")
            regkeys = odm.Optional(odm.List(odm.SSDeepHash(copyto="__text__")), description="Registry Keys")

        @odm.model(index=True, store=False, description="Windows")
        class DynamicWindow(odm.Model):
            cls_ids = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="CLSIDs")
            dynamic_classes = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Dynamic Classes")
            regkeys = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Registry Keys")

        @odm.model(index=True, store=False, description="Operating System")
        class DynamicOperatingSystem(odm.Model):
            platform = odm.Optional(odm.List(odm.Platform(copyto="__text__")), description="Platform")
            version = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Version")
            processor = odm.Optional(odm.List(odm.Processor(copyto="__text__")), description="Processor")

        autorun_location = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Autorun location")
        dos_device = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="DOS Device")
        mutex = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Mutex")
        registry_key = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Registy Keys")
        process = odm.Optional(odm.Compound(DynamicProcess), description="Sandbox Processes")
        signature = odm.Optional(odm.Compound(DynamicSignature), description="Sandbox Signatures")
        ssdeep = odm.Optional(odm.Compound(DynamicSSDeep), description="Sandbox SSDeep")
        window = odm.Optional(odm.Compound(DynamicWindow), description="Sandbox Window")
        operating_system = odm.Optional(odm.Compound(DynamicOperatingSystem), description="Sandbox Operating System")
        processtree_id = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Process Tree ID")

    @odm.model(index=True, store=False, description="General Information Tag Model")
    class Info(odm.Model):
        phone_number = odm.Optional(odm.List(odm.PhoneNumber(copyto="__text__")))
        password = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Password")

    @odm.model(index=True, store=False, description="File Tag Model")
    class File(odm.Model):
        @odm.model(index=True, store=False, description="APK File Model")
        class FileAPK(odm.Model):
            @odm.model(index=True, store=False, description="APK Application Model")
            class FileAPKApp(odm.Model):
                label = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Label")
                version = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Version")

            @odm.model(index=True, store=False, description="APK SDK Model")
            class FileAPKSDK(odm.Model):
                min = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Minimum OS required")
                target = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Target OS")

            activity = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Activity")
            app = odm.Optional(odm.Compound(FileAPKApp), description="APK Application Information")
            feature = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Features")
            locale = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Locale")
            permission = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Permissions")
            pkg_name = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Package Name")
            provides_component = odm.Optional(odm.List(odm.Keyword(copyto="__text__")),
                                              description="Components Provided")
            sdk = odm.Optional(odm.Compound(FileAPKSDK), description="APK SDK Information")
            used_library = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Libraries Used")

        @odm.model(index=True, store=False, description="File Date Model")
        class FileDate(odm.Model):
            creation = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="File Creation Date")
            last_modified = odm.Optional(odm.List(odm.Keyword(copyto="__text__")),
                                         description="File Last Modified Date")

        @odm.model(index=True, store=False, description="ELF File Tag Model")
        class FileELF(odm.Model):
            @odm.model(index=True, store=False, description="ELF Sections")
            class FileELFSections(odm.Model):
                name = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Section Name")

            @odm.model(index=True, store=False, description="ELF Segments")
            class FileELFSegments(odm.Model):
                type = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Segment Type")

            @odm.model(index=True, store=False, description="ELF Notes")
            class FileELFNotes(odm.Model):
                name = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Note Name")
                type = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Note Type")
                type_core = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Note Core Type")

            libraries = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Libraries")
            interpreter = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Interpreter")
            sections = odm.Optional(odm.Compound(FileELFSections), description="ELF Sections")
            segments = odm.Optional(odm.Compound(FileELFSegments), description="ELF Segments")
            notes = odm.Optional(odm.Compound(FileELFNotes), description="ELF Notes")

        @odm.model(index=True, store=False, description="Image File Tag Model")
        class FileIMG(odm.Model):
            @odm.model(index=True, store=False, description="Exiftool Information Model")
            class FileIMGExiftool(odm.Model):
                creator_tool = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Image Creation Tool")
                derived_document_id = odm.Optional(odm.List(odm.Keyword(copyto="__text__")),
                                                   description="Derived Document ID")
                document_id = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Document ID")
                instance_id = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Instance ID")
                toolkit = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Toolkit")

            exif_tool = odm.Optional(odm.Compound(FileIMGExiftool), description="Exiftool Information")
            mega_pixels = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Megapixels")
            mode = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Image Mode")
            size = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Image Size")
            sorted_metadata_hash = odm.Optional(odm.List(odm.Keyword(copyto="__text__")),
                                                description="Sorted Metadata Hash")

        @odm.model(index=True, store=False, description="JAR File Tag Model")
        class FileJAR(odm.Model):
            main_class = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Main Class")
            main_package = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Main Package")

        @odm.model(index=True, store=False, description="File Name Model")
        class FileName(odm.Model):
            anomaly = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Name of Anomaly")
            extracted = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Name of Extracted")

        @odm.model(index=True, store=False, description="OLE File Tag Model")
        class FileOLE(odm.Model):
            @odm.model(index=True, store=False, description="OLE Macro Model")
            class FileOLEMacro(odm.Model):
                sha256 = odm.Optional(odm.List(odm.SHA256(copyto="__text__")), description="SHA256 of Macro")
                suspicious_string = odm.Optional(odm.List(odm.Keyword(copyto="__text__")),
                                                 description="Suspicious Strings")

            @odm.model(index=True, store=False, description="OLE Summary Model")
            class FileOLESummary(odm.Model):
                author = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Author")
                codepage = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Code Page")
                comment = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Comment")
                company = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Company")
                create_time = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Creation Time")
                last_printed = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Date Last Printed")
                last_saved_by = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="User Last Saved By")
                last_saved_time = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Date Last Saved")
                manager = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Manager")
                subject = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Subject")
                title = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Title")

            macro = odm.Optional(odm.Compound(FileOLEMacro), description="OLE Macro")
            summary = odm.Optional(odm.Compound(FileOLESummary), description="OLE Summary")
            clsid = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="CLSID")
            dde_link = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="DDE Link")
            fib_timestamp = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="FIB Timestamp")

        @odm.model(index=True, store=False, description="PDF File Tag Model")
        class FilePDF(odm.Model):
            @odm.model(index=True, store=False, description="PDF Date Model")
            class FilePDFDate(odm.Model):
                modified = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Date Modified")
                pdfx = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="PDFx")
                source_modified = odm.Optional(odm.List(odm.Keyword(copyto="__text__")),
                                               description="Date Source Modified")

            @odm.model(index=True, store=False, description="PDF Javascript Model")
            class FilePDFJavascript(odm.Model):
                sha1 = odm.Optional(odm.List(odm.SHA1(copyto="__text__")), description="SHA1 of Javascript")

            @odm.model(index=True, store=False, description="PDF Statistics Model")
            class FilePDFStats(odm.Model):
                sha1 = odm.Optional(odm.List(odm.SHA1(copyto="__text__")), description="SHA1 of Statistics")

            date = odm.Optional(odm.Compound(FilePDFDate), description="PDF Date Information")
            javascript = odm.Optional(odm.Compound(FilePDFJavascript), description="PDF Javascript Information")
            stats = odm.Optional(odm.Compound(FilePDFStats), description="PDF Statistics Information")

        @odm.model(index=True, store=False, description="PE File Tag Model")
        class FilePE(odm.Model):
            @odm.model(index=True, store=False, description="PE Debug Model")
            class FilePEDebug(odm.Model):
                guid = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="GUID")

            @odm.model(index=True, store=False, description="PE Exports Model")
            class FilePEExports(odm.Model):
                function_name = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Function Name")
                module_name = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Module Name")

            @odm.model(index=True, store=False, description="PE Imports Model")
            class FilePEImports(odm.Model):
                fuzzy = odm.Optional(odm.List(odm.SSDeepHash(copyto="__text__")), description="Fuzzy")
                md5 = odm.Optional(odm.List(odm.MD5(copyto="__text__")), description="MD5")
                imphash = odm.Optional(odm.List(odm.MD5(copyto="__text__")), description="Imphash")
                sorted_fuzzy = odm.Optional(odm.List(odm.SSDeepHash(copyto="__text__")), description="Sorted Fuzzy")
                sorted_sha1 = odm.Optional(odm.List(odm.SHA1(copyto="__text__")), description="Sorted SHA1")
                gimphash = odm.Optional(odm.List(odm.SHA256(copyto="__text__")), description="Go Import hash")
                suspicious = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Suspicious")

            @odm.model(index=True, store=False, description="PE Linker Model")
            class FilePELinker(odm.Model):
                timestamp = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Timestamp")

            @odm.model(index=True, store=False, description="PE OEP Model")
            class FilePEOEP(odm.Model):
                bytes = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Bytes")
                hexdump = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Hex Dump")

            @odm.model(index=True, store=False, description="PE Resources Model")
            class FilePEResources(odm.Model):
                language = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Language")
                name = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Name")

            @odm.model(index=True, store=False, description="PE Rich Header Model")
            class FilePERichHeader(odm.Model):
                hash = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Hash")

            @odm.model(index=True, store=False, description="PE Sections Model")
            class FilePESections(odm.Model):
                hash = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Hash")
                name = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Name")

            @odm.model(index=True, store=False, description="PE Versions Model")
            class FilePEVersions(odm.Model):
                description = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Description")
                filename = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Filename")

            api_vector = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="API Vector")
            debug = odm.Optional(odm.Compound(FilePEDebug), description="PE Debug Information")
            exports = odm.Optional(odm.Compound(FilePEExports), description="PE Exports Information")
            imports = odm.Optional(odm.Compound(FilePEImports), description="PE Imports Information")
            linker = odm.Optional(odm.Compound(FilePELinker), description="PE Linker Information")
            oep = odm.Optional(odm.Compound(FilePEOEP), description="PE OEP Information")
            pdb_filename = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="PDB Filename")
            resources = odm.Optional(odm.Compound(FilePEResources), description="PE Resources Information")
            rich_header = odm.Optional(odm.Compound(FilePERichHeader), description="PE Rich Header Information")
            sections = odm.Optional(odm.Compound(FilePESections), description="PE Sections Information")
            versions = odm.Optional(odm.Compound(FilePEVersions), description="PE Versions Information")

        @odm.model(index=True, store=False, description="PList File Tag Model")
        class FilePList(odm.Model):
            @odm.model(index=True, store=False, description="PList Build Model")
            class FilePListBuild(odm.Model):
                machine_os = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Machine OS")

            @odm.model(index=True, store=False, description="PList CF Bundle Model")
            class FilePListCFBundle(odm.Model):
                @odm.model(index=True, store=False, description="PList CF Bundle Version Model")
                class FilePListCFBundleVersion(odm.Model):
                    long = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Long Version")
                    short = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Short Version")

                development_region = odm.Optional(odm.List(odm.Keyword(copyto="__text__")),
                                                  description="Development Region")
                display_name = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Display Name")
                executable = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Executable Name")
                identifier = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Identifier Name")
                name = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Bundle Name")
                pkg_type = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Package Type")
                signature = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Signature")
                url_scheme = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="URL Scheme")
                version = odm.Optional(odm.Compound(FilePListCFBundleVersion), description="Bundle Version Information")

            @odm.model(index=True, store=False, description="PList DT Model")
            class FilePListDT(odm.Model):
                @odm.model(index=True, store=False, description="PList DT Platform Model")
                class FilePListDTPlatform(odm.Model):
                    build = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Build")
                    name = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Name")
                    version = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Version")

                compiler = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Compiler")
                platform = odm.Optional(odm.Compound(FilePListDTPlatform), description="Platform Information")

            @odm.model(index=True, store=False, description="PList LS Model")
            class FilePListLS(odm.Model):
                background_only = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Background Only")
                min_system_version = odm.Optional(odm.List(odm.Keyword(copyto="__text__")),
                                                  description="Minimum System Versuion")

            @odm.model(index=True, store=False, description="PList NS Model")
            class FilePListNS(odm.Model):
                apple_script_enabled = odm.Optional(odm.List(odm.Keyword(copyto="__text__")),
                                                    description="Apple Script Enabled")
                principal_class = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Principal Class")

            @odm.model(index=True, store=False, description="PList UI Model")
            class FilePListUI(odm.Model):
                background_modes = odm.Optional(
                    odm.List(odm.Keyword(copyto="__text__")),
                    description="Background Modes")
                requires_persistent_wifi = odm.Optional(odm.List(odm.Keyword(copyto="__text__")),
                                                        description="Requires Persistent WIFI")

            @odm.model(index=True, store=False, description="PList WK Model")
            class FilePListWK(odm.Model):
                app_bundle_identifier = odm.Optional(odm.List(odm.Keyword(copyto="__text__")),
                                                     description="App Bundle ID")

            installer_url = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Installer URL")
            min_os_version = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Minimum OS Version")
            requests_open_access = odm.Optional(odm.List(odm.Keyword(copyto="__text__")),
                                                description="Requests Open Access")

            build = odm.Optional(odm.Compound(FilePListBuild), description="Build Information")
            cf_bundle = odm.Optional(odm.Compound(FilePListCFBundle), description="CF Bundle Information")
            dt = odm.Optional(odm.Compound(FilePListDT), description="DT Information")
            ls = odm.Optional(odm.Compound(FilePListLS), description="LS Information")
            ns = odm.Optional(odm.Compound(FilePListNS), description="NS Information")
            ui = odm.Optional(odm.Compound(FilePListUI), description="UI Information")
            wk = odm.Optional(odm.Compound(FilePListWK), description="WK Information")

        @odm.model(index=True, store=False, description="PowerShell File Tag Model")
        class FilePowerShell(odm.Model):
            cmdlet = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Cmdlet")

        @odm.model(index=True, store=False, description="Shortcut File Tag Model")
        class FileShortcut(odm.Model):
            command_line = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Command Line")
            icon_location = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Icon Location")
            machine_id = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Machine ID")
            tracker_mac = odm.Optional(
                odm.List(odm.Keyword(copyto="__text__")), description="Possible MAC address from the Tracker block"
            )

        @odm.model(index=True, store=False, description="Strings File Model")
        class FileStrings(odm.Model):
            api = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="API")
            blacklisted = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Blacklisted")
            decoded = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Decoded")
            extracted = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Extracted")

        @odm.model(index=True, store=False, description="SWF File Tag Model")
        class FileSWF(odm.Model):
            @odm.model(index=True, store=False, description="SWF Header Model")
            class FileSWFHeader(odm.Model):
                @odm.model(index=True, store=False, description="SWF Header Frame")
                class FileSWFHeaderFrame(odm.Model):
                    count = odm.Optional(odm.List(odm.Integer()), description="Number of Frames")
                    rate = odm.Optional(odm.List(odm.Keyword()), description="Speed of Animation")
                    size = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Size of Frame")

                frame = odm.Optional(odm.Compound(FileSWFHeaderFrame), description="Header Frame Information")
                version = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Version")

            header = odm.Optional(odm.Compound(FileSWFHeader), description="Header Information")
            tags_ssdeep = odm.Optional(odm.List(odm.SSDeepHash(copyto="__text__")), description="Tags SSDeep")

        behavior = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="File Behaviour")
        compiler = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Compiler of File")
        config = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="File Configuration")
        date = odm.Optional(odm.Compound(FileDate), description="File's Date Information")
        elf = odm.Optional(odm.Compound(FileELF), description="ELF File Properties")
        lib = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="File Libraries")
        name = odm.Optional(odm.Compound(FileName), description="File Name")
        path = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="File Path")
        rule = odm.Optional(odm.Mapping(odm.List(odm.Keyword(copyto="__text__"))), description="Rule/Signature File")
        string = odm.Optional(odm.Compound(FileStrings), description="File Strings Properties")
        apk = odm.Optional(odm.Compound(FileAPK), description="APK File Properties")
        jar = odm.Optional(odm.Compound(FileJAR), description="JAR File Properties")
        img = odm.Optional(odm.Compound(FileIMG), description="Image File Properties")
        ole = odm.Optional(odm.Compound(FileOLE), description="OLE File Properties")
        pe = odm.Optional(odm.Compound(FilePE), description="PE File Properties")
        pdf = odm.Optional(odm.Compound(FilePDF), description="PDF File Properties")
        plist = odm.Optional(odm.Compound(FilePList), description="PList File Properties")
        powershell = odm.Optional(odm.Compound(FilePowerShell), description="PowerShell File Properties")
        shortcut = odm.Optional(odm.Compound(FileShortcut), description="Shortcut File Properties")
        swf = odm.Optional(odm.Compound(FileSWF), description="SWF File Properties")

    @odm.model(index=True, store=False, description="Network Tag Model")
    class Network(odm.Model):
        @odm.model(index=True, store=False, description="Network IOC Model")
        class NetworkIOCs(odm.Model):
            domain = odm.Optional(odm.List(odm.Domain(copyto="__text__")), description="Domain")
            ip = odm.Optional(odm.List(odm.IP(copyto="__text__")), description="IP")
            uri = odm.Optional(odm.List(odm.URI(copyto="__text__")), description="URI")
            uri_path = odm.Optional(odm.List(odm.URIPath(copyto="__text__")), description="URI Path")

        @odm.model(index=True, store=False, description="Network Email Model")
        class NetworkEmail(odm.Model):
            address = odm.Optional(odm.List(odm.Email(copyto="__text__")), description="Email Address")
            date = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Date")
            subject = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Subject")
            msg_id = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Message ID")

        @odm.model(index=True, store=False, description="Network Signature Model")
        class NetworkSignature(odm.Model):
            signature_id = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Signature ID")
            message = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Signature Message")

        @odm.model(index=True, store=False, description="Network TLS Model")
        class NetworkTLS(odm.Model):
            ja3_hash = odm.Optional(odm.List(odm.MD5(copyto="__text__")), description="JA3 Hash")
            ja3_string = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="JA3 String")
            sni = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="SNI")

        attack = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Attack")
        dynamic = odm.Optional(odm.Compound(NetworkIOCs), description="Dynamic IOCs")
        email = odm.Optional(odm.Compound(NetworkEmail), description="Email")
        mac_address = odm.Optional(odm.List(odm.MAC(copyto="__text__")), description="MAC Address")
        port = odm.Optional(odm.List(odm.Integer()), description="Port")
        protocol = odm.Optional(odm.List(odm.Keyword()), description="Protocol")
        signature = odm.Optional(odm.Compound(NetworkSignature), description="Signatures")
        static = odm.Optional(odm.Compound(NetworkIOCs), description="Static IOCs")
        tls = odm.Optional(odm.Compound(NetworkTLS), description="TLS")
        user_agent = odm.Optional(odm.List(odm.Keyword()), description="User Agent")

    @odm.model(index=True, store=False, description="Technique Tag Model")
    class Technique(odm.Model):
        comms_routine = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Communication Routine")
        config = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Configuration")
        crypto = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Cryptography")
        keylogger = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Keylogger")
        macro = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Macro")
        masking_algo = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Masking Algorithm")
        obfuscation = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Obfuscation")
        packer = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Packer")
        persistence = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Persistence")
        shellcode = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Shell Code")
        string = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="String")

    attribution = odm.Optional(odm.Compound(Attribution), description="Attribution Tagging")
    av = odm.Optional(odm.Compound(AV), description="Antivirus Tagging")
    cert = odm.Optional(odm.Compound(Cert), description="Certificate Tagging")
    dynamic = odm.Optional(odm.Compound(Dynamic), description="Dynamic Analysis Tagging")
    info = odm.Optional(odm.Compound(Info), description="Informational Tagging")
    file = odm.Optional(odm.Compound(File), description="File Tagging")
    network = odm.Optional(odm.Compound(Network), description="Network Tagging")
    source = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Source Tagging")
    technique = odm.Optional(odm.Compound(Technique), description="Technique Tagging")
    vector = odm.Optional(odm.List(odm.Keyword(copyto="__text__")), description="Vector Tagging")


if __name__ == "__main__":
    from pprint import pprint

    pprint(list(Tagging().flat_fields().keys()))



from assemblyline import odm
from assemblyline.odm.models.ontology.results.process import ObjectID
from assemblyline.common.dict_utils import get_dict_fingerprint_hash

OID_PARTS = ['engine_name', 'virus_name']
TAG_PARTS = ['engine_name', 'virus_name']


@odm.model(description="Antivirus Ontology Model")
class Antivirus(odm.Model):
    objectid = odm.Compound(ObjectID, description="The object ID of the antivirus object")
    engine_name = odm.Keyword(description="Name of antivirus engine")
    engine_version = odm.Optional(odm.Keyword(), description="Version of antivirus engine")
    engine_definition_version = odm.Optional(odm.Keyword(), description="Version of definition set")
    virus_name = odm.Optional(odm.Keyword(), description="The name of the virus")
    # What category does the verdict fall under?
    category = odm.Optional(odm.Enum(['type-unsupported',
                                      'undetected',
                                      'failure',
                                      'suspicious',
                                      'malicious']),
                            description="What category does the verdict fall under?<br><ul>"
                            "<li>`type-unsupported`: File sent to antivirus is unsupported</li>"
                            "<li>`undetected`: File not detected by antivirus</li>"
                            "<li>`failure`: Antivirus failed during detection</li>"
                            "<li>`suspicious`: Antivirus deems suspicious</li>"
                            "<li>`malicious`: Antivirus deems malicious</li></ul>")

    def get_oid(data: dict):
        return f"antivirus_{get_dict_fingerprint_hash({key: data.get(key) for key in OID_PARTS})}"

    def get_tag(data: dict):
        return ".".join([data.get(key) for key in TAG_PARTS if data.get(key)])
